import 'package:flutter/cupertino.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_post/ocg_post.dart';
import 'package:ocg_post/src/models/post.dart';
import 'package:ocg_post/src/repository/post/base.dart';

abstract class PostListEvent {}

class PostListRefreshEvent extends PostListEvent {}

class PostListAddAllEvent extends PostListEvent {
  final List<Post> posts;

  PostListAddAllEvent(this.posts);
}

class PostListReactEvent extends PostListEvent {
  final String postId;
  final UserReaction userReaction;

  PostListReactEvent(this.postId, this.userReaction);
}

class PostListReactDeleteEvent extends PostListEvent {
  final String postId;
  final UserReaction userReaction;

  PostListReactDeleteEvent(this.postId, this.userReaction);
}

abstract class PostListState {}

class PostListInitialState extends PostListState {}

class PostListLoadingState extends PostListState {}

class PostListLoadedState extends PostListState {
  final Map<String, Post> posts;
  PostListLoadedState(this.posts);
}

class PostListErrorState extends PostListState {
  final String error;
  PostListErrorState(this.error);
}

class PostListBloc extends Bloc<PostListEvent, PostListState> {
  Map<String, Post> _posts = Map<String, Post>();
  final PostRepository postRepository;
  PostListBloc(this.postRepository) : super(PostListInitialState());
  @override
  Stream<PostListState> mapEventToState(PostListEvent event) async* {
    if (event is PostListRefreshEvent) {
      yield PostListLoadingState();
      AppResponse<List<Post>> response = await postRepository.getPage(null, 10);
      if (response.isSuccess) {
        response.data.forEach((element) {
          _posts[element.id] = element;
        });
        yield PostListLoadedState(_posts);
      } else {
        yield PostListErrorState(response.error);
      }
    }
    if (event is PostListAddAllEvent) {
      var posts = event.posts;
      posts.forEach((element) {
        _posts[element.id] = element;
      });
      yield PostListLoadedState(_posts);
    }
    if (event is PostListReactEvent) {
      AppResponse<bool> response =
          await postRepository.reactToPost(event.postId, event.userReaction);
      if (response.isSuccess) {
        if (_posts[event.postId] != null) {
          Post post = _posts[event.postId];
          post = post.copyWith(reactionCount: (post.reactionCount ?? 0) + 1);
          var myReactions = (post.myReactions ?? []).toList();
          if (myReactions.indexWhere(
                  (element) => element.type == event.userReaction.type) <
              0) {
            myReactions.add(event.userReaction);
          }
          post = post.copyWith(myReactions: myReactions);
          _posts[event.postId] = post;
          yield PostListLoadedState(_posts);
        }
      } else {}
    }
    if (event is PostListReactDeleteEvent) {
      AppResponse<bool> response = await postRepository.deleteReactToPost(
          event.postId, event.userReaction);
      if (response.isSuccess) {
        if (_posts[event.postId] != null) {
          Post post = _posts[event.postId];
          post = post.copyWith(
              reactionCount:
                  (post.reactionCount ?? 0) > 0 ? post.reactionCount - 1 : 0);
          var myReactions = (post.myReactions ?? []).toList();
          if (myReactions.indexWhere(
                  (element) => element.type == event.userReaction.type) >=
              0) {
            myReactions.remove(event.userReaction);
          }
          post = post.copyWith(myReactions: myReactions);
          _posts[event.postId] = post;
          yield PostListLoadedState(_posts);
        }
      } else {}
    }
  }

  List<Post> sorted(List<Post> posts) {
    var list = posts.toList();
    list.sort((a, b) => b.recordedAt.compareTo(a.recordedAt));
    return list;
  }
}
