import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_post/src/models/comment.dart';
import 'package:ocg_post/src/repository/comment/base.dart';
import 'package:ocg_profile/ocg_profile.dart';

class CommentFirebaseRepository extends CommentRepository {
  final String postCollectionName;
  final ProfileRepository profileRepository;
  final TokenRepository tokenRepository;

  CommentFirebaseRepository(
      this.tokenRepository, this.profileRepository, this.postCollectionName);

  CollectionReference commentCollection(String postId) =>
      FirebaseFirestore.instance
          .collection(postCollectionName)
          .doc(postId)
          .collection("comments");

  @override
  Future<AppResponse<String>> create(Comment object) async {
    try {
      assert(object.postId != null);
      assert(object.description != null);
      var tokenResponse = await tokenRepository.getToken();
      if (tokenResponse.isSuccess) {
        String userId = tokenResponse.data.userId;
        assert(userId != null);
        object = object.copyWith(
          profile: Profile.named(id: userId),
          recordedAt: DateTime.now(),
          id: IDHelper().generateId(),
        );
        var json = toJson(object);
        await commentCollection(object.postId).doc(object.id).set(json);
        return AppResponse.named(data: object.id);
      } else {
        return AppResponse.named(error: tokenResponse.error);
      }
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Comment>>> getPage(
      Comment comment, String postId, int count) async {
    try {
      QuerySnapshot snapshot;
      if (comment != null) {
        snapshot = await commentCollection(postId)
            .orderBy("recordedAt", descending: true)
            .startAfter([comment.recordedAt.toIso8601String()])
            .limit(count)
            .get();
      } else {
        snapshot = await commentCollection(postId)
            .orderBy("recordedAt", descending: true)
            .limit(count)
            .get();
      }
      if (snapshot != null) {
        var result = await Future.wait<Comment>(snapshot.docs.map((e) async {
          var response = await fromJson(e.data());
          return response;
        }));
        result = result.where((element) => element != null).toList();
        return AppResponse.named(data: result);
      }
      return AppResponse.named(error: "Some error occurred");
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  Map<String, dynamic> toJson(Comment comment) {
    var json = comment.toJson();
    json["profileId"] = comment?.profile?.id;
    return json;
  }

  Future<Comment> fromJson(var json) async {
    var profileId = json["profileId"];
    if (profileId != null) {
      var response = await profileRepository.read(profileId);
      if (response.isSuccess) json["profile"] = response.data.toJson();
    }
    Comment comment = Comment.fromJson(json);
    var tokenResponse = await tokenRepository.getToken();
    if (tokenResponse.isSuccess) {
      comment =
          comment.copyWith(isSelf: profileId == tokenResponse.data.userId);
    }
    return comment;
  }

  @override
  Future<AppResponse<bool>> delete(String postId, String commentId) async {
    try {
      var tokenResponse = await tokenRepository.getToken();
      var response = await read(postId, commentId);
      if (response.isSuccess &&
          response.data?.profile?.id == tokenResponse?.data?.userId) {
        await commentCollection(postId).doc(commentId).delete();
        return AppResponse.named(data: true);
      }
      return AppResponse.named(error: "Comment can not be deleted");
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<Comment>> read(String postId, String commentId) async {
    try {
      var snapshot = await commentCollection(postId).doc(commentId).get();
      var comment = await fromJson(snapshot.data());
      return AppResponse.named(data: comment);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }
}
