import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/foundation.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_location/ocg_location.dart';
import 'package:ocg_post/ocg_post.dart';
import 'package:ocg_post/src/models/post.dart';
import 'package:ocg_post/src/repository/post/base.dart';
import 'package:ocg_post/src/repository/reaction/firebase.dart';
import 'package:ocg_profile/ocg_profile.dart';

class PostFirebaseRepository extends PostRepository {
  final FirebaseFirestore _firestore;
  final IDHelper idHelper;
  final ProfileRepository profileRepository;
  final HashtagRepository hashtagRepository;
  final PostReactionFirebaseRepository reactionRepository;
  final TokenRepository tokenRepository;
  final CommentRepository commentRepository;

  static const String posts = 'posts';

  CollectionReference get postCollection => _firestore.collection(posts);

  PostFirebaseRepository(this.idHelper, this.tokenRepository,
      this.profileRepository, this.hashtagRepository, this.reactionRepository,
      {this.commentRepository})
      : this._firestore = FirebaseFirestore.instance;

  @override
  Future<AppResponse<String>> create(
    Post object,
  ) async {
    try {
      AppResponse<String> userResponse = await profileRepository.getCurrentId();
      if (userResponse.isSuccess) {
        String userId = userResponse.data;
        String id = idHelper.generateId();
        List<Hashtag> hashtagsCreated = [];
        hashtagsCreated = List<Hashtag>.from(
            await Future.wait((object.hashtags ?? []).map((e) async {
          AppResponse<String> response = await hashtagRepository.create(e);
          if (response.isSuccess) {
            String id = response.data;
            return e.copyWith(id: id);
          }
        }).toList()));
        hashtagsCreated =
            hashtagsCreated.where((element) => element != null).toList();
        object = object.copyWith(
            id: id,
            hashtags: hashtagsCreated,
            profile: Profile.named(id: userId),
            recordedAt: DateTime.now());
        var json = _toJson(object);
        await postCollection.doc(id).set(json);
        return AppResponse.named(data: id);
      } else {
        return AppResponse.named(error: "Not logged in");
      }
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  Map<String, dynamic> _toJson(Post post) {
    var json = post.toJson();
    json["hashtags"] = post.hashtags.map((e) => e.id).toList();
    json["profileId"] = post?.profile?.id;
    json["profile"] = null;
    if (post.taggedPeople != null) {
      var map = {};
      post.taggedPeople.forEach((element) {
        map[element.id] = true;
      });
      json["taggedPeople"] = map;
    }
    return json;
  }

  Post _fromJson(Map json) {
    // Hashtags
    var hashtagMap = json["hashtags"];
    var newHashtags = [];
    if (hashtagMap != null && hashtagMap is List) {
      newHashtags =
          hashtagMap.map((e) => Hashtag.named(id: e).toJson()).toList();
    }
    json["hashtags"] = newHashtags;
    try {
      var reactionMap = json["reactions"];
      var reactions = [];
      int count = 0;
      if (reactionMap != null && reactionMap is Map) {
        reactions = reactionMap.keys
            .map((e) => PostReaction.named(
                    type: e, count: (reactionMap[e] ?? {})["count"] ?? 0)
                .toJson())
            .toList();
        reactions.forEach((element) {
          count += element["count"];
        });
      }
      json["reactions"] = reactions;
      json["reactionCount"] = count;
    } catch (e) {
      debugPrint("Error $e");
    }
    return Post.fromJson(json);
  }

  @override
  Future<AppResponse<bool>> delete(String id) async {
    try {
      var response = await read(id);
      if (response.isSuccess && response.data.isSelf) {
        await postCollection.doc(id).delete();
        return AppResponse.named(data: true);
      }
      return AppResponse.named(error: "Post not deleted");
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Post>>> getList(
      {int page = 1, int count = 10}) async {
    var snapshot = await postCollection.limit(count * page).get();
    var documents =
        snapshot.docs.map((e) => e.data()).skip((page - 1) * count).toList();
    var list = (await Future.wait(documents.map((e) async {
      return await _parseJson(e);
    }).toList()))
        .where((element) => element != null)
        .toList();
    return AppResponse.named(data: list);
  }

  Future<Post> _parseJson(postJson) async {
    var newJson = postJson ?? {};
    if (postJson["profileId"] is String) {
      var profileResponse = await profileRepository.read(postJson['profileId']);
      if (profileResponse.isSuccess) {
        postJson["profile"] = profileResponse.data.toJson();
      }
      var taggedPeople = postJson["taggedPeople"];
      if (taggedPeople is Map) {
        List<String> ids = List<String>.from(taggedPeople.keys.toList());
        var profiles = await Future.wait<Profile>(ids.map((e) async {
          var response = await profileRepository.read(e);
          if (response.isSuccess) {
            return response.data;
          }
          return null;
        }).toList());
        postJson["taggedPeople"] = profiles.map((e) => e.toJson()).toList();
      }
      await Future.wait([
        Future(() async {
          var reactionResponse =
              await reactionRepository.getMyReactions(postJson["id"]);
          if (reactionResponse.isSuccess) {
            newJson["myReactions"] =
                reactionResponse.data.map((e) => e.toJson()).toList();
          }
        }),
        Future(() async {
          if (commentRepository == null) {
            return;
          }
          var response =
              await commentRepository.getPage(null, postJson["id"], 2);
          if (response.isSuccess) {
            newJson["topComments"] =
                response.data.map((e) => e.toJson()).toList();
          } else {
            newJson["topComments"] = [];
          }
        })
      ]);
      var tokenResponse = await tokenRepository.getToken();
      var post = _fromJson(newJson);
      post =
          post.copyWith(isSelf: tokenResponse.data.userId == post.profile?.id);
      return post;
    }
    return null;
  }

  @override
  Future<AppResponse<Post>> read(String id) async {
    try {
      var document = await postCollection.doc(id).get();
      var data = document.data();
      return AppResponse.named(data: await _parseJson(data));
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> update(Post object) async {
    try {
      await postCollection.doc(object.id).update(object.toJson());
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Post>>> getFollowingPosts(
      {int count = 10, int page = 1, Post after}) async {
    try {
      var currentResponse = await profileRepository.getCurrentId();
      if (!currentResponse.isSuccess) {
        return AppResponse.named(error: "${currentResponse.error}");
      }
      String currentId = currentResponse.data;
      AppResponse<List<ProfileFollowing>> profileResponse =
          await profileRepository.getFollowersFollowing(currentId);

      if (profileResponse.isSuccess) {
        List<String> following = List<String>.from(profileResponse.data
            .where((element) => element.from == currentId)
            .map((e) => e.to)
            .toList());
        var query = postCollection.orderBy("recordedAt", descending: true);
        if (after != null) {
          query = query.where("recordedAt",
              isLessThanOrEqualTo: after.recordedAt.toIso8601String());
        }
        var snapshot = await query
            .where("profileId", whereIn: (following ?? [])..add(currentId))
            .limit(count * page)
            .get();
        var documents = snapshot.docs
            .map((e) => e.data())
            .skip((page - 1) * count)
            .toList();
        var list = (await Future.wait(documents.map((e) async {
          return await _parseJson(e);
        }).toList()))
            .where((element) => element != null && element.id != after?.id)
            .toList();
        return AppResponse.named(data: list);
      } else {
        return AppResponse.named(error: profileResponse.error);
      }
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Post>>> getNearbyPosts(Position position,
      {int count = 10, Post after}) async {
    try {
      RangeGeoHash rangeGeoHash = LocationRepository.getGeoHashRange(
          position.latitude, position.longitude);
      var query = postCollection
          .limit(count)
          .where(
            "taggedLocation.geohash",
            isGreaterThanOrEqualTo: rangeGeoHash.lower,
          )
          .where("taggedLocation.geohash",
              isLessThanOrEqualTo: rangeGeoHash.upper);

      // if (after != null) {
      //   query = query.startAfter([after.recordedAt.toIso8601String()]);
      // }
      var snapshot = await query.get();
      var documents = snapshot.docs.map((e) => e.data()).toList();
      var list = (await Future.wait(documents.map((e) async {
        return await _parseJson(e);
      }).toList()))
          .where((element) => element != null)
          .toList();
      return AppResponse.named(data: list);
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<bool>> reactToPost(
      String postId, UserReaction userReaction) async {
    var response = await reactionRepository.reactToPost(postId, userReaction);
    return response;
  }

  @override
  Future<AppResponse<bool>> deleteReactToPost(
      String postId, UserReaction userReaction) async {
    var response =
        await reactionRepository.deleteReactToPost(postId, userReaction);
    return response;
  }

  @override
  Future<AppResponse<List<Post>>> getProfilePosts(String profileId,
      {int count = 10, int page = 1}) async {
    try {
      var snapshot =
          await postCollection.where("profileId", isEqualTo: profileId).get();
      var docs = snapshot.docs.map((e) => e.data()).toList();
      var posts = List<Post>.from(await Future.wait(docs.map((e) async {
        return await _parseJson(e);
      }).toList()));
      posts = posts.where((element) => element != null).toList();
      return AppResponse.named(data: posts);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Post>>> getHashtagPosts(Hashtag hashtag,
      {int count = 10, int page = 1}) async {
    try {
      var snapshot = await postCollection
          .where("hashtags", arrayContains: hashtag.id)
          .orderBy("recordedAt", descending: true)
          .limit(count * page)
          .get();
      var docs = snapshot.docs.map((e) => e.data()).toList();
      var posts = List<Post>.from(await Future.wait(docs.map((e) async {
        return await _parseJson(e);
      }).toList()));
      posts = posts.where((element) => element != null).toList();
      return AppResponse.named(data: posts);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<Post>>> getPage(Post object, int count) {
    // TODO: implement getPage
    throw UnimplementedError();
  }
}
