import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/foundation.dart';
import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_post/src/models/post.dart';
import 'package:ocg_post/src/repository/reaction/base.dart';
import 'package:ocg_profile/ocg_profile.dart';

class PostReactionFirebaseRepository extends PostReactionRepository {
  final String postCollectionName;
  final ProfileRepository profileRepository;
  final TokenRepository tokenRepository;
  PostReactionFirebaseRepository(this.tokenRepository, this.profileRepository,
      {this.postCollectionName});

  String get reactions => "reactions";
  String get reactionCount => "reactionCount";

  CollectionReference reactionCollection(String postId) =>
      postCollection.doc(postId).collection(reactions);
  CollectionReference get postCollection =>
      FirebaseFirestore.instance.collection(postCollectionName);

  @override
  Future<AppResponse<List<UserReaction>>> getMyReactions(String postId) async {
    try {
      var tokenResponse = await tokenRepository.getToken();
      if (tokenResponse.isSuccess) {
        var userId = tokenResponse.data?.userId;
        var snapshot = await reactionCollection(postId)
            .where(
              "userId",
              isEqualTo: userId,
            )
            .get();
        var docs =
            snapshot.docs.map((e) => UserReaction.fromJson(e.data())).toList();
        return AppResponse.named(data: List<UserReaction>.from(docs));
      }
      return AppResponse.named(error: tokenResponse.error);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<bool>> deleteMyReaction(
      String postId, UserReaction userReaction) async {
    try {
      assert(userReaction.type != null);
      var tokenResponse = await tokenRepository.getToken();
      if (tokenResponse.isSuccess) {
        var userId = tokenResponse.data?.userId;
        var snapshot = await reactionCollection(postId)
            .where(
              "userId",
              isEqualTo: userId,
            )
            .where("type", isEqualTo: userReaction.type)
            .get();
        if (snapshot.docs.length > 0) {
          await Future.wait([
            Future(() async {
              // React
              await reactionCollection(postId)
                  .doc(snapshot.docs[0].id)
                  .delete();
            }),
            Future(() async {
              // React
              await postCollection.doc(postId).update(
                  {"${userReaction.type}.count": FieldValue.increment(-1)});
            }),
          ]);
          return AppResponse.named(data: true);
        } else {
          return AppResponse.named(data: false);
        }
      }
      return AppResponse.named(error: "tokenResponse.error");
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<PostReaction>>> getPostReactions(
      String postId) async {
    if (postCollection == null) {
      return AppResponse.named(error: "Not implemented");
    }
    try {
      var snapshot = await postCollection.doc(postId).get();
      var doc = snapshot.data();
      var reactions = doc["reactions"];
      if (reactions is Map) {
        var list = reactions.keys
            .map((e) => PostReaction.named(
                type: e, count: (reactions[e] ?? {})["count"] ?? 0))
            .toList();
        return AppResponse.named(data: List<PostReaction>.from(list));
      }
      return AppResponse.named(data: []);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<UserReaction>>> getReactions(
    String postId,
  ) async {
    try {
      var snapshot = await reactionCollection(postId).get();
      var docs = await Future.wait<UserReaction>(
          snapshot.docs.map((e) async => await fromJson(e.data())).toList());
      docs = docs.where((element) => element != null).toList();

      return AppResponse.named(data: List<UserReaction>.from(docs));
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<List<UserReaction>>> getReactionsByType(
      String postId, String type) async {
    try {
      var snapshot =
          await reactionCollection(postId).where("type", isEqualTo: type).get();
      var docs =
          snapshot.docs.map((e) => UserReaction.fromJson(e.data())).toList();
      return AppResponse.named(data: List<UserReaction>.from(docs));
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<bool>> reactToPost(
      String postId, UserReaction userReaction) async {
    assert(userReaction.type != null);
    try {
      var tokenResponse = await tokenRepository.getToken();
      if (!tokenResponse.isSuccess) {
        return AppResponse.named(error: tokenResponse.error);
      }
      String userId = tokenResponse.data.userId;
      // Check for already reacted
      var s1 = await reactionCollection(postId)
          .where("type", isEqualTo: userReaction.type)
          .where("userId", isEqualTo: userId)
          .get();
      if (s1.docs.length > 0) {
        return AppResponse.named(error: "Already Reacted");
      }
      await Future.wait([
        Future(() async {
          // React
          userReaction = userReaction.copyWith(
              profile: Profile.named(id: userId),
              id: IDHelper().generateId(),
              recordedAt: DateTime.now());
          await reactionCollection(postId)
              .doc(userReaction.id)
              .set(toJson(userReaction));
        }),
        Future(() async {
          // React
          await postCollection.doc(postId).update({
            "reactions.${userReaction.type}.count": FieldValue.increment(1)
          });
        }),
      ]);
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  @override
  Future<AppResponse<bool>> deleteReactToPost(
      String postId, UserReaction userReaction) async {
    assert(userReaction.type != null);
    try {
      var tokenResponse = await tokenRepository.getToken();
      if (!tokenResponse.isSuccess) {
        return AppResponse.named(error: tokenResponse.error);
      }
      // Check for already reacted
      await Future.wait([
        Future(() async {
          await reactionCollection(postId).doc(userReaction.id).delete();
        }),
        Future(() async {
          await postCollection.doc(postId).update({
            "reactions.${userReaction.type}.count": FieldValue.increment(-1)
          });
        }),
      ]);
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: "$e");
    }
  }

  Map<String, dynamic> toJson(UserReaction reaction) {
    var toJson = reaction.toJson();
    toJson["profile"] = null;
    toJson["userId"] = reaction.profile?.id;
    return toJson;
  }

  Future<UserReaction> fromJson(Map<String, dynamic> json) async {
    try {
      var userId = json["userId"];
      var response = await profileRepository.read(userId);
      json["profile"] = response?.data?.toJson();
      return UserReaction.fromJson(json);
    } catch (e) {
      return null;
    }
  }
}
